@@ -112,8 +112,123 @@ gss_release_name (OM_uint32 * minor_status, gss_name_t * name)
   return GSS_S_COMPLETE;
 }
 
-/* See meta.c for gss_inquire_names_for_mech() and
-   gss_inquire_mechs_for_name() */
+OM_uint32
+gss_inquire_names_for_mech (OM_uint32 * minor_status,
+			    const gss_OID mechanism, gss_OID_set * name_types)
+{
+  OM_uint32 maj_stat;
+  _gss_mech_api_t mech;
+  int i;
+
+  mech = _gss_find_mech (mechanism);
+
+  maj_stat = gss_create_empty_oid_set (minor_status, name_types);
+  if (maj_stat != GSS_S_COMPLETE)
+    return maj_stat;
+
+  for (i = 0; mech->name_types[i]; i++)
+    {
+      maj_stat = gss_add_oid_set_member (minor_status, mech->name_types[i],
+					 name_types);
+      if (maj_stat != GSS_S_COMPLETE)
+	{
+	  gss_release_oid_set (minor_status, name_types);
+	  return maj_stat;
+	}
+    }
+
+  if (minor_status)
+    *minor_status = 0;
+  return GSS_S_COMPLETE;
+}
+
+static OM_uint32
+_gss_inquire_mechs_for_name2 (OM_uint32 * minor_status,
+			      _gss_mech_api_t mech,
+			      gss_OID name_type,
+			      gss_OID_set *mech_types)
+{
+  gss_OID_set oids;
+  int supported;
+  OM_uint32 maj_stat;
+
+  maj_stat = gss_inquire_names_for_mech (minor_status, mech->mech, &oids);
+  if (maj_stat != GSS_S_COMPLETE)
+    return maj_stat;
+
+  maj_stat = gss_test_oid_set_member (minor_status, name_type,
+				      oids, &supported);
+  if (maj_stat != GSS_S_COMPLETE)
+    {
+      gss_release_oid_set (minor_status, &oids);
+      return maj_stat;
+    }
+
+  if (supported)
+    {
+      maj_stat = gss_add_oid_set_member (minor_status,
+					 mech->mech,
+					 mech_types);
+      if (maj_stat != GSS_S_COMPLETE)
+	{
+	  gss_release_oid_set (minor_status, &oids);
+	  return maj_stat;
+	}
+    }
+  gss_release_oid_set (minor_status, &oids);
+
+  if (minor_status)
+    *minor_status = 0;
+  return GSS_S_COMPLETE;
+}
+
+static OM_uint32
+_gss_inquire_mechs_for_name1 (OM_uint32 * minor_status,
+			      gss_OID name_type,
+			      gss_OID_set *mech_types)
+{
+  OM_uint32 maj_stat;
+  int i;
+
+  for (i = 0; _gss_mech_apis[i].mech; i++)
+    {
+      maj_stat = _gss_inquire_mechs_for_name2 (minor_status,
+					       &_gss_mech_apis[i],
+					       name_type,
+					       mech_types);
+      if (maj_stat != GSS_S_COMPLETE)
+	return maj_stat;
+    }
+
+  if (minor_status)
+    *minor_status = 0;
+  return GSS_S_COMPLETE;
+}
+
+OM_uint32
+gss_inquire_mechs_for_name (OM_uint32 * minor_status,
+			    const gss_name_t input_name,
+			    gss_OID_set * mech_types)
+{
+  int i;
+  OM_uint32 maj_stat;
+
+  maj_stat = gss_create_empty_oid_set (minor_status, mech_types);
+  if (maj_stat != GSS_S_COMPLETE)
+    return maj_stat;
+
+  maj_stat = _gss_inquire_mechs_for_name1 (minor_status, input_name->type,
+					   mech_types);
+  if (maj_stat != GSS_S_COMPLETE)
+    {
+      gss_release_oid_set (minor_status, mech_types);
+      return maj_stat;
+    }
+
+  if (minor_status)
+    *minor_status = 0;
+  return GSS_S_COMPLETE;
+}
 
 OM_uint32
 gss_export_name (OM_uint32 * minor_status,
